# A module for examining numeric values, allowing a peek into their binary
# representations within the CPU, the make-up of Float values, and a list of
# values of other types which share the same bit pattern.
#
# IEEE 754 is the international standard that defines how float values are
# represented within a fixed number of bits - the Wikipedia entry for
# IEEE754 is a good place to start learning the meaning of the various
# components of a float value.
#
# All methods are module methods that should be called with the module as the
# explicit receiver, and come in both 32bit (single precision) and 64bit (double
# precision) versions.
#
#   IEEE754.to_s_32(-0.1, "My Test Value")
#
#   #=> IEEE 32 BIT: My Test Value
#   #=> Float -0.10000000149011612
#   #=> Signed -1110651699 : Unsigned 3184315597
#   #=> Hex Bytes 3D,CC,CC,CD
#   #=> Binary 10111101,11001100,11001100,11001101
#   #=> Sign 1 : Expo 01111011 : Frac 10011001100110011001101
#   #=>    -ve :       (2**-4) * 1.6000000238418579
#
# Free for all to use, distribute and/or modify.  There is no warranty, not even
# an implied warranty that the module is fit for purpose (caveat emptor!).
#
# v1.0.0, June 2014, Trog Luddite
#
module IEEE754

  extend(self)

  # Take the given number, which may be a Float, Integer or String containing a
  # numeric value, ensure that it is a 32 bit value, and display...
  # - The float number represented by the value's bit pattern; i.e. the actual
  #   value stored in memory when precision is truncated to 32 bits.  The number
  #   is shown with 64bit precision to give an idea when a 32bit float is
  #   not exactly representing the decimal input value.
  # - The signed and unsigned integers that share the same bit pattern.
  # - The sequence of hex bytes making up the value (MSB first)
  # - A breakdown of the IEEE float components that define the value.
  # - Warnings of infinities, NaNs and denormal values.
  # An optional second argument allows a String label to be added to the
  # output to help identify the value being inspected.
  #
  def to_s_32(num, msg = nil)
    if num.is_a?(String)
      num = (num =~ /\./) ? Float(num) : Integer(num)
    end    
    case num
    when Integer
      f32 = int32_to_float(num)
      i32 = single_to_int32(f32)
      u32 = single_to_uint32(f32)
    when Float
      u32 = single_to_uint32(num)
      i32 = single_to_int32(num)
      f32 = int32_to_float(i32)
    else
      raise(ArgumentError, "Numeric value expected", caller)
    end
    hex = ('%08X' % u32).scan(/../).join(',')
    bin = '%032b' % u32
    sign, expo, frac = bin[0], bin[1..8], bin[9..31]
    str = 'IEEE 32 BIT' << (msg ? ": #{msg}" : '')
    str << "\r\nFloat #{f32}\r\nSigned #{i32} : Unsigned #{u32}\r\nHex Bytes #{hex}"
    str << "\r\nBinary #{bin.scan(/.{8}/).join(',')}"
    str << "\r\nSign %1s : Expo %8s : Frac %23s" % [sign, expo, frac]
    expo = expo.to_i(2) - 127
    unless expo == 128
      sign = (sign == '0') ? '+ve' : '-ve'
      pow  = "(2**#{expo}) * "
      mant = (expo == -127 ? '0' : '1') + (frac.to_i(2) * 2**-23).to_f.to_s[1..-1]
      str << "\r\n   %s : %16s%s" % [sign, pow, mant]
      str << ' DENORMAL!' if expo == -127  
    end
    str
  end

  # Take the given number, which may be a Float, Integer or String containing a
  # numeric value.  Ensure that this is a 64bit value, and then display...
  # - The float number represented by the value's bit pattern; note that the
  #   displayed number may not be completely accurate for decimal values with
  #   no binary equivalent (standard Ruby cannot display decimals with any
  #   greater precision).
  # - The signed and unsigned integers that share the same bit pattern.
  # - The sequence of hex bytes making up the value (MSB first)
  # - A breakdown of the IEEE float components that define the value.
  # - Warnings of infinities, NaNs and denormal values.
  # An optional second argument allows a String label to be added to the
  # output to help identify the value being inspected.
  #
  def to_s_64(num, msg = nil)
    if num.is_a?(String)
      num = (num =~ /\./) ? Float(num) : Integer(num)
    end 
    case num
    when Integer
      f64 = int64_to_float(num)
      i64 = double_to_int64(f64)
      u64 = double_to_uint64(f64)
    when Float
      u64 = double_to_uint64(num)
      i64 = double_to_int64(num)
      f64 = int64_to_float(i64)
    else
      raise(ArgumentError, "Numeric value expected", caller)
    end
    hex = ('%016X' % u64).scan(/../).join(',')
    bin = '%064b' % u64
    sign, expo, frac = bin[0], bin[1..11], bin[12..63]
    str = 'IEEE 64 BIT' << (msg ? ": #{msg}" : '')
    str << "\r\nFloat #{f64}\r\nSigned #{i64} : Unsigned #{u64}\r\nHex Bytes #{hex}"
    str << "\r\nBinary #{bin.scan(/.{8}/).join(',')}"
    str << "\r\nSign %1s : Expo %11s : Frac %52s" % [sign, expo, frac]
    expo = expo.to_i(2) - 1023
    unless expo == 1024
      sign = (sign == '0') ? '+ve' : '-ve'
      pow  = "(2**#{expo}) * "
      mant = (expo == -1023 ? '0' : '1') + (frac.to_i(2) * 2**-52).to_f.to_s[1..-1]
      str << "\r\n   %s : %16s%s" % [sign, pow, mant]
      str << ' DENORMAL!' if expo == -1023 
    end
    str
  end

  # Return the signed Integer value that shares the same bit-pattern as the
  # given Float.  Floats are truncated to the precision (no. of bits) first.
  def single_to_int32(float) ; [float].pack('F').unpack('l')[0] ; end
  def double_to_int64(float) ; [float].pack('D').unpack('q')[0] ; end
  
  # Return the unsigned Integer value that shares the same bit-pattend as the
  # given Float.  Floats are truncated to the precision (no. of bits) first.
  def single_to_uint32(float) ; [float].pack('F').unpack('L')[0] ; end
  def double_to_uint64(float) ; [float].pack('D').unpack('Q')[0] ; end
  
  # Return the Float that shares the same bit pattern as the given integer.
  # Integers are truncated to the correct number of bits first.
  def int32_to_float(int) ; [int].pack('L').unpack('F')[0] ; end
  def int64_to_float(int) ; [int].pack('Q').unpack('D')[0] ; end
  
end # module

## TEST ##
# puts IEEE754.to_s_32(-0.15625)  # -1.25 * 2**-3
# puts
# puts IEEE754.to_s_64(-0.15625)  # -1.25 * 2**-3
# puts
# puts IEEE754.to_s_32("-0.1")    # Shows rounding to 32 bit precision
# puts
# puts IEEE754.to_s_64("-0.1")
# puts
# puts IEEE754.to_s_32(1_000_000_003.0)  # Shows rounding to 32 bit precision
# puts
# puts IEEE754.to_s_64(-1.0/0.0, "-INF?")  # Minus Infinity
# puts
# puts IEEE754.to_s_32(0x0000003A, "Denorm?")  # Denormal value
# puts
# puts IEEE754.to_s_32(0xFFF00001, "NaN?")  # Not a Number!!
